import time
import contextlib


@contextlib.contextmanager
def timed_block():
    """
    A context manager that times the block of code within it. The elapsed
    time will be updated in the `elapsed` attribute of the returned object
    until the block is exited.

    Example:

    .. code-block:: python

        with timed_block() as timer:
            time.sleep(1)

        print(f"Block took {timer.elapsed} seconds.")
    """

    class Timer:
        def __init__(self):
            self.start = time.monotonic()
            self.stop = None

        @property
        def elapsed(self):
            return (self.stop or time.monotonic()) - self.start

    timer = Timer()
    yield timer
    timer.stop = time.monotonic()


def importable_name(obj):
    """
    Get the importable name for an object.

    .. note::

        This will not work for objects that are not defined in a module,
        such as lambdas, REPL functions, functions-in-functions, etc...

    :param obj: The object to get the importable name for.
    :return: str
    """
    return f"{obj.__module__}.{obj.__qualname__}"



from psycopg import AsyncConnection
from psycopg import sql

from chancy.app import Chancy
from chancy.worker import Worker
from chancy.utils import timed_block
from chancy.plugin import Plugin, PluginScope


class Recovery(Plugin):
    """
    Recovers jobs that appear to be abandoned by a worker.

    Typically, this happens when a worker is unexpectedly terminated, or has
    otherwise been lost which we recognize by checking the last seen timestamp
    of the worker heartbeat.

    This will transition any matching jobs back to the "pending" state, and
    increment the `max_attempts` counter by 1 to allow it to be retried.

    Events
    ------

    The following events are emitted by the recovery plugin:

    +---------------------+------------------------------------------------+
    | Event Name          | Description                                    |
    +=====================+================================================+
    | `recovery.recovered`| Emitted when the recovery has recovered jobs.  |
    +---------------------+------------------------------------------------+

    :param poll_interval: The number of seconds between recovery poll intervals.
    """

    def __init__(self, *, poll_interval: int = 60):
        super().__init__()
        self.poll_interval = poll_interval

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker: Worker, chancy: Chancy):
        while await self.sleep(self.poll_interval):
            await self.wait_for_leader(worker)
            async with chancy.pool.connection() as conn:
                with timed_block() as chancy_time:
                    rows_recovered = await self.recover(worker, chancy, conn)
                    self.log.info(
                        f"Recovery recovered {rows_recovered} row(s) from the"
                        f" database. Took {chancy_time.elapsed:.2f} seconds."
                    )
                    await worker.hub.emit(
                        "recovery.recovered",
                        elapsed=chancy_time.elapsed,
                        rows_recovered=rows_recovered,
                    )

    async def recover(
        self, worker: Worker, chancy: Chancy, conn: AsyncConnection
    ) -> int:
        """
        Recover jobs that were running when the worker was unexpectedly
        terminated, or has otherwise been lost.

        :param worker: The worker that is running the recovery.
        :param chancy: The Chancy application.
        :param conn: The database connection.
        :return: The number of rows recovered from the database
        """
        query = sql.SQL(
            """
            UPDATE
                {jobs} cj
            SET
                state = 'pending',
                taken_by = NULL,
                started_at = NULL,
                max_attempts = max_attempts + 1
            WHERE
               NOT EXISTS (
                    SELECT 1
                    FROM {workers} cw
                    WHERE (
                        cw.worker_id = cj.taken_by
                        AND
                        cw.last_seen >= NOW() - INTERVAL '{interval} SECOND'
                    )
              )
              AND state = 'running';
        """
        ).format(
            jobs=sql.Identifier(f"{chancy.prefix}jobs"),
            workers=sql.Identifier(f"{chancy.prefix}workers"),
            interval=sql.Literal(worker.heartbeat_timeout),
        )

        async with conn.cursor() as cursor:
            async with conn.transaction():
                await cursor.execute(query)
                return cursor.rowcount



from psycopg import AsyncConnection
from psycopg import sql

from chancy.migrate import Migration, Migrator


class V1Migration(Migration):
    async def up(self, migrator: Migrator, conn: AsyncConnection):
        """
        Create the initial $prefix_jobs, $prefix_leaders tables.
        """
        async with conn.transaction():
            await conn.execute(
                sql.SQL(
                    """
                    CREATE TABLE {jobs} (
                        id BIGSERIAL PRIMARY KEY,
                        queue TEXT NOT NULL,
                        payload JSON NOT NULL,
                        state VARCHAR (25) NOT NULL DEFAULT 'pending',
                        priority INTEGER DEFAULT 10,
                        attempts INTEGER DEFAULT 0,
                        max_attempts INTEGER DEFAULT 1,
                        taken_by TEXT,
                        created_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                        started_at TIMESTAMPTZ,
                        completed_at TIMESTAMPTZ,
                        scheduled_at TIMESTAMPTZ NOT NULL DEFAULT NOW()
                    )
                    """
                ).format(jobs=sql.Identifier(f"{migrator.prefix}jobs"))
            )

            await conn.execute(
                sql.SQL(
                    """
                    CREATE UNLOGGED TABLE {leader} (
                        id BIGSERIAL PRIMARY KEY,
                        worker_id TEXT NOT NULL,
                        first_seen TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                        expires_at TIMESTAMPTZ NOT NULL
                    );
                    ALTER TABLE {leader}
                        ADD CONSTRAINT leader_worker_id_unique UNIQUE
                            (worker_id);
                    """
                ).format(leader=sql.Identifier(f"{migrator.prefix}leader"))
            )

            await conn.execute(
                sql.SQL(
                    """
                    CREATE UNLOGGED TABLE {workers} (
                        worker_id TEXT NOT NULL PRIMARY KEY,
                        last_seen TIMESTAMPTZ NOT NULL DEFAULT NOW()
                    );
                    """
                ).format(workers=sql.Identifier(f"{migrator.prefix}workers"))
            )

    async def down(self, migrator: Migrator, conn: AsyncConnection):
        """
        Drop the $prefix_jobs, $prefix_leaders tables.
        """
        async with conn.transaction():
            await conn.execute(
                sql.SQL("DROP TABLE {jobs}").format(
                    jobs=sql.Identifier(f"{migrator.prefix}jobs")
                )
            )
            await conn.execute(
                sql.SQL("DROP TABLE {leader}").format(
                    leaders=sql.Identifier(f"{migrator.prefix}leader")
                )
            )
            await conn.execute(
                sql.SQL("DROP TABLE {workers}").format(
                    workers=sql.Identifier(f"{migrator.prefix}workers")
                )
            )



import abc
import json
import asyncio
from typing import Callable
from functools import cached_property

from psycopg import sql, AsyncConnection, AsyncCursor
from psycopg.rows import dict_row
from psycopg.types.json import Json

from chancy.executor import Executor, JobInstance, Limit, Job
from chancy.executors.process import ProcessExecutor
from chancy.plugin import Plugin, PluginScope


class QueuePlugin(Plugin, abc.ABC):
    """
    A specialized plugin that provides a queue for pull jobs from, and clients
    to push jobs to.

    This class is an abstract base class that should be subclassed to implement
    a queue provider, such as a Redis-backed queue or an SQS-backed queue.

    :param name: The name of the queue.
    :param concurrency: The maximum number of jobs that can be run concurrently
                        by each worker on this queue.
    :param executor: The executor to use for running jobs.
    :param polling_interval: The interval at which to poll the queue for new
                             jobs.
    """

    def __init__(
        self,
        name: str,
        *,
        concurrency: int = 1,
        executor: Callable[["QueuePlugin"], Executor] | None = None,
        polling_interval: int | None = 5,
    ):
        super().__init__()
        self.name = name
        self.concurrency = concurrency
        self.polling_interval = polling_interval
        self._executor = ProcessExecutor if executor is None else executor
        # A queue of pending updates to jobs that need to be applied on the
        # next fetch.
        self.pending_updates = asyncio.Queue()

    @cached_property
    def executor(self) -> Executor:
        return self._executor(self)

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.QUEUE

    @abc.abstractmethod
    async def push(self, app, jobs: list[Job]):
        """
        Push one or more jobs onto the queue.

        :param app: The Chancy application.
        :param jobs: The jobs to push onto the queue.
        """


class Queue(QueuePlugin):
    """
    A postgres-backed queue that uses the database to store jobs.
    """

    def __init__(
        self,
        name: str,
        *,
        concurrency: int = 1,
        executor: Callable[["Queue"], Executor] | None = None,
        polling_interval: int | None = 5,
    ):
        super().__init__(
            name,
            concurrency=concurrency,
            executor=executor,
            polling_interval=polling_interval,
        )
        # A queue of pending updates to jobs that need to be applied on the
        # next fetch.
        self.pending_updates = asyncio.Queue()

    async def run(self, worker, app):
        """
        Continuously polls the queue for new jobs.

        This method will run indefinitely, polling the queue for new jobs and
        running them as they become available.

        .. note::

            If you want to pull jobs from the queue to process them yourself,
            you can use the :meth:`fetch()` method directly.

        :param app: The app that is polling the queue.
        :param worker: The worker that is polling the queue.
        """
        while await self.sleep(self.polling_interval):
            async with app.pool.connection() as conn:
                # If we wouldn't be able to run a job even if we had one, we
                # should just wait. Pre-fetching can be advantageous, but
                # IMO it causes more headache (as seen with Celery and future
                # scheduled tasks) than it's worth.
                maximum_jobs_to_poll = self.concurrency - len(self.executor)
                if maximum_jobs_to_poll <= 0:
                    self.log.debug("No capacity for new jobs, skipping poll.")
                    await asyncio.sleep(self.polling_interval)
                    continue

                jobs = await self.fetch_jobs(
                    conn,
                    up_to=maximum_jobs_to_poll,
                    prefix=app.prefix,
                    worker_id=worker.worker_id,
                )
                await worker.hub.emit(
                    "queue.polled",
                    fetched=len(jobs),
                    worker_id=worker.worker_id,
                    queue=self.name,
                )

                for job in jobs:
                    self.log.debug(f"Found job {job.id}, pushing to executor.")
                    await self.executor.push(job)

    async def fetch_jobs(
        self,
        conn: AsyncConnection,
        *,
        up_to: int = 1,
        prefix: str = "chancy_",
        worker_id: str | None = None,
    ) -> list[JobInstance]:
        """
        Fetch jobs from the queue.

        This method will fetch up to `up_to` jobs from the queue, mark them as
        running, and return them as a list of `JobInstance` objects.

        It's safe to call this method concurrently, as the jobs will be locked
        for the duration of the transaction.

        :param conn: The database connection to use.
        :param up_to: The maximum number of jobs to fetch.
        :param prefix: The prefix to use for the database tables.
        :param worker_id: The ID of the worker fetching the jobs.
        """
        jobs_table = sql.Identifier(f"{prefix}jobs")

        async with conn.cursor(row_factory=dict_row) as cursor:
            async with conn.transaction():
                if not self.pending_updates.empty():
                    # Gather all pending updates and apply them to the jobs
                    # table before fetching new jobs.
                    pending_updates = []
                    while not self.pending_updates.empty():
                        pending_updates.append(await self.pending_updates.get())

                    try:
                        await cursor.executemany(
                            sql.SQL(
                                """
                                UPDATE
                                    {jobs}
                                SET
                                    state = %(state)s,
                                    completed_at = %(completed_at)s,
                                    attempts = %(attempts)s
                                WHERE
                                    id = %(id)s
                                """
                            ).format(jobs=jobs_table),
                            [
                                {
                                    "id": update.id,
                                    "state": update.state,
                                    "completed_at": update.completed_at,
                                    "attempts": update.attempts,
                                }
                                for update in pending_updates
                            ],
                        )
                    except Exception:
                        # If we were unable to apply the updates, we should
                        # re-queue them for the next poll.
                        for update in pending_updates:
                            await self.pending_updates.put(update)
                        raise

                await cursor.execute(
                    sql.SQL(
                        """
                        WITH selected_jobs AS (
                            SELECT
                                id
                            FROM
                                {jobs}
                            WHERE
                                queue = %(queue)s
                            AND
                                (state = 'pending' or state = 'retrying')
                            AND
                                attempts < max_attempts
                            AND
                                (scheduled_at IS NULL OR scheduled_at <= NOW())
                            ORDER BY
                                priority ASC,
                                created_at ASC,
                                id ASC
                            LIMIT
                                %(maximum_jobs_to_fetch)s
                            FOR UPDATE OF {jobs} SKIP LOCKED
                        )
                        UPDATE
                            {jobs}
                        SET
                            started_at = NOW(),
                            attempts = attempts + 1,
                            state = 'running',
                            taken_by = %(worker_id)s
                        FROM
                            selected_jobs
                        WHERE
                            {jobs}.id = selected_jobs.id
                        RETURNING {jobs}.*
                        """
                    ).format(
                        jobs=jobs_table,
                    ),
                    {
                        "queue": self.name,
                        "maximum_jobs_to_fetch": up_to,
                        "worker_id": worker_id,
                    },
                )

                records = await cursor.fetchall()

            return [
                JobInstance(
                    id=record["id"],
                    func=record["payload"]["func"],
                    kwargs=record["payload"]["kwargs"],
                    priority=record["priority"],
                    scheduled_at=record["scheduled_at"],
                    created_at=record["created_at"],
                    started_at=record["started_at"],
                    completed_at=record["completed_at"],
                    attempts=record["attempts"],
                    max_attempts=record["max_attempts"],
                    state=record["state"],
                    limits=[
                        Limit.deserialize(limit)
                        for limit in record["payload"]["limits"]
                    ],
                )
                for record in records
            ]

    async def push_job_update(self, job: JobInstance):
        """
        Push an update to a job.

        This method is typically called by an executor after a job has been
        completed (either successfully or with an error) to update the job's
        state in the database. It does _not_ immediately update the job,
        instead they are batched and updated the next time the worker polls
        the queue.
        """
        await self.pending_updates.put(job)

    async def push(self, app, jobs: list[Job]):
        async with app.pool.connection() as conn:
            async with conn.cursor() as cursor:
                async with conn.transaction():
                    await self.push_jobs(cursor, jobs, prefix=app.prefix)

    async def push_jobs(
        self, cursor: AsyncCursor, jobs: list[Job], prefix: str = "chancy_"
    ):
        """
        Push one or more jobs onto the queue.

        This non-standard queue method can be used to push jobs to the database
        within an existing transaction. This is very useful when you want to
        push jobs to the queue only if related objects are successfully
        created, like running onboarding for a new user only if the user is
        successfully created.

        :param cursor: The database cursor to use.
        :param jobs: The jobs to push onto the queue.
        :param prefix: The prefix to use for the database tables.
        """
        await cursor.executemany(
            sql.SQL(
                """
                INSERT INTO
                    {jobs} (
                        queue,
                        payload,
                        priority,
                        max_attempts,
                        scheduled_at
                ) VALUES (%s, %s, %s, %s, %s);
                """
            ).format(jobs=sql.Identifier(f"{prefix}jobs")),
            [
                (
                    self.name,
                    Json(
                        {
                            "func": job.func,
                            "kwargs": job.kwargs or {},
                            "limits": [
                                limit.serialize() for limit in job.limits
                            ],
                        }
                    ),
                    job.priority,
                    job.max_attempts,
                    job.scheduled_at,
                )
                for job in jobs
            ],
        )
        await cursor.execute(
            sql.SQL("SELECT pg_notify({events}, {event});").format(
                events=sql.Literal(f"{prefix}events"),
                event=sql.Literal(json.dumps({"t": "pushed", "q": self.name})),
            )
        )



from psycopg import AsyncConnection
from psycopg import sql

from chancy.app import Chancy
from chancy.worker import Worker
from chancy.plugin import Plugin, PluginScope
from chancy.plugins.rule import RuleT, AgeRule
from chancy.utils import timed_block


class Pruner(Plugin):
    """
    A plugin that prunes stale data from the database.

    Jobs
    ----

    The pruner will never prune jobs that haven't been run yet ("pending"),
    or are currently being run ("running").

    You can use simple rules, or combine them using the `|` and `+` operators
    to create complex rules.

    For example, to prune jobs that are older than 60 seconds:

    .. code-block:: python

        from chancy.plugins.rule import AgeRule
        Pruner(AgeRule(60))

    Or to prune jobs that are older than 60 seconds and are in the "default"
    queue:

    .. code-block:: python

        from chancy.plugins.rule import AgeRule, QueueRule
        Pruner(QueueRule("default") + AgeRule(60))

    Or to prune jobs that are older than 60 seconds and are in the "default"
    queue, or instantly deleted if the job is `update_cache`:

    .. code-block:: python

        from chancy.plugins.rule import AgeRule, QueueRule, JobRule
        Pruner((QueueRule("default") + AgeRule(60)) | JobRule("update_cache"))

    By default, the pruner will run every 60 seconds and will remove up to
    10,000 jobs in a single run that have been completed for more than 60
    seconds.

    .. note::

        By default, only an AgeRule will be covered by an index. If you use
        multiple rules, you may need to create additional indexes to improve
        performance on busy queues.

    Events
    ------

    The following events are emitted by the pruner plugin:

    +---------------------+------------------------------------------------+
    | Event Name          | Description                                    |
    +=====================+================================================+
    | `pruner.removed`    | Emitted when the pruner has removed jobs from  |
    |                     | the database.                                  |
    +---------------------+------------------------------------------------+

    :param rule: The rule that the pruner will use to match jobs.
    :param maximum_to_prune: The maximum number of jobs to prune in a single
                             run of the pruner.
    :param poll_interval: The interval in seconds between each run of the
                          pruner.
    """

    def __init__(
        self,
        rule: RuleT = AgeRule(60),
        *,
        maximum_to_prune: int = 10000,
        poll_interval: int = 60,
    ):
        super().__init__()
        self.rule = rule
        self.maximum_to_prune = maximum_to_prune
        self.poll_interval = poll_interval

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker: Worker, chancy: Chancy):
        while await self.sleep(self.poll_interval):
            await self.wait_for_leader(worker)
            async with chancy.pool.connection() as conn:
                with timed_block() as chancy_time:
                    rows_removed = await self.prune(worker, chancy, conn)
                    self.log.info(
                        f"Pruner removed {rows_removed} row(s) from the"
                        f" database. Took {chancy_time.elapsed:.2f} seconds."
                    )
                    await worker.hub.emit(
                        "pruner.removed",
                        elapsed=chancy_time.elapsed,
                        rows_removed=rows_removed,
                    )

    async def prune(
        self, worker: Worker, chancy: Chancy, conn: AsyncConnection
    ) -> int:
        """
        Prune stale records from the database.

        :param worker: The worker that is running the pruner.
        :param chancy: The Chancy application.
        :param conn: The database connection.
        :return: The number of rows removed from the database
        """
        job_query = sql.SQL(
            """
            WITH jobs_to_prune AS (
                SELECT ctid
                FROM {table}
                WHERE state NOT IN ('pending', 'running')
                AND ({rule})
                LIMIT {maximum_to_prune}
            )
            DELETE FROM {table}
            WHERE ctid IN (SELECT ctid FROM jobs_to_prune)
            """
        ).format(
            table=sql.Identifier(f"{chancy.prefix}jobs"),
            rule=self.rule.to_sql(),
            maximum_to_prune=sql.Literal(self.maximum_to_prune),
        )

        async with conn.cursor() as cursor:
            async with conn.transaction():
                await cursor.execute(job_query)
                return cursor.rowcount



import abc
import enum
import typing
import dataclasses
from datetime import datetime, timezone
from typing import Any, Optional

from chancy.utils import importable_name

if typing.TYPE_CHECKING:
    from chancy.queue import QueuePlugin


@dataclasses.dataclass
class Limit:
    """
    A limit that can be applied to a job.
    """

    class Type(enum.Enum):
        MEMORY = "memory"
        TIME = "time"

    type_: "Type"
    value: int

    @classmethod
    def deserialize(cls, data: dict) -> "Limit":
        return cls(type_=cls.Type(data["t"]), value=data["v"])

    def serialize(self) -> dict:
        return {"t": self.type_.value, "v": self.value}


@dataclasses.dataclass(frozen=True, kw_only=True)
class Job:
    """
    A job is an immutable, stateless unit of work that can be pushed onto a
    Chancy queue and executed elsewhere.
    """

    func: str
    kwargs: dict[str, Any] | None = dataclasses.field(default_factory=dict)
    priority: int = 0
    max_attempts: int = 1
    scheduled_at: datetime = dataclasses.field(
        default_factory=lambda: datetime.now(tz=timezone.utc)
    )
    limits: list[Limit] = dataclasses.field(default_factory=list)

    @classmethod
    def from_func(cls, func, **kwargs):
        return cls(func=importable_name(func), **kwargs)

    def with_priority(self, priority: int) -> "Job":
        return dataclasses.replace(self, priority=priority)

    def with_max_attempts(self, max_attempts: int) -> "Job":
        return dataclasses.replace(self, max_attempts=max_attempts)

    def with_scheduled_at(self, scheduled_at: datetime) -> "Job":
        return dataclasses.replace(self, scheduled_at=scheduled_at)

    def with_limits(self, limits: list[Limit]) -> "Job":
        return dataclasses.replace(self, limits=limits)

    def with_kwargs(self, kwargs: dict[str, Any]) -> "Job":
        return dataclasses.replace(self, kwargs=kwargs)


@dataclasses.dataclass(frozen=True, kw_only=True)
class JobInstance(Job):
    """
    A job instance is a job that has been pushed onto a queue and now has
    stateful information associated with it, such as the number of attempts
    so far.
    """

    class State:
        PENDING = "pending"
        RUNNING = "running"
        FAILED = "failed"
        RETRYING = "retrying"
        SUCCEEDED = "succeeded"

    id: int
    created_at: datetime
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    attempts: int = 0
    state: State = State.PENDING


class Executor(abc.ABC):
    """
    The executor is responsible for managing the execution of jobs in a job
    pool.
    """

    def __init__(self, queue: "QueuePlugin"):
        self.queue = queue

    @abc.abstractmethod
    async def push(self, job: JobInstance):
        """
        Push a job onto the job pool.
        """

    @abc.abstractmethod
    def __len__(self):
        """
        Get the number of pending and running jobs in the pool.
        """



import asyncio
import dataclasses
import functools
import os
import resource
import signal
import traceback
import typing
import threading
from asyncio import Future
from concurrent.futures import ProcessPoolExecutor
from typing import Callable

from chancy.executor import Executor, JobInstance, Limit


class _TimeoutThread(threading.Thread):
    """
    A thread that will raise a TimeoutError after a specified number of
    seconds.
    """

    def __init__(self, timeout: int, cancel: threading.Event):
        super().__init__()
        self.timeout = timeout
        self.cancel = cancel

    def run(self):
        if self.cancel.wait(self.timeout) is True:
            # Our call to wait() returning True means that the flag was set
            # before the timeout elapsed, so we should cancel our alarm.
            return

        # If we reach this point, the timeout has elapsed, and we should raise
        # a TimeoutError back in the main process thread.
        os.kill(os.getpid(), signal.SIGALRM)


class ProcessExecutor(Executor):
    """
    An Executor which uses a process pool to run its jobs.

    This executor is useful for running jobs that are CPU-bound, avoiding the
    GIL (Global Interpreter Lock) that Python uses to ensure thread safety.

    If a timeout is requested for a job, a separate thread will be spawned
    within each process to raise a TimeoutError if the job takes too long to
    complete.

    .. note::

        Currently, this is the default executor used by Chancy.

    :param queue: The queue that this executor is associated with.
    :param maximum_jobs_per_worker: The maximum number of jobs that each worker
                                    can run before being replaced.
    """

    def __init__(self, queue, *, maximum_jobs_per_worker: int = 100):
        super().__init__(queue)

        self.processes: dict[Future, JobInstance] = {}
        self.pool = ProcessPoolExecutor(
            max_workers=self.queue.concurrency,
            max_tasks_per_child=maximum_jobs_per_worker,
        )

    async def push(self, job: JobInstance) -> Future:
        future: Future = self.pool.submit(self.job_wrapper, job)
        future.add_done_callback(
            functools.partial(
                self.job_completed, loop=asyncio.get_running_loop()
            )
        )
        self.processes[future] = job
        return future

    def __len__(self):
        return len(self.processes)

    @classmethod
    def job_wrapper(cls, job: JobInstance):
        """
        This is the function that is actually started by the process pool
        executor. It's responsible for setting up necessary signals and limits,
        running the job, and returning the result.

        Subclasses can override this method to provide additional functionality
        or to change the way that jobs are run.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """
        cleanup: list[Callable] = []

        for limit in job.limits:
            match limit.type_:
                case Limit.Type.TIME:
                    # We need this to work on platforms that don't support
                    # SIGALRM (looking at you, Windows). So we use a separate
                    # thread to raise a TimeoutError if the job takes too long.
                    signal.signal(signal.SIGALRM, cls.job_signal_handler)
                    cancel = threading.Event()
                    timeout_thread = _TimeoutThread(limit.value, cancel)
                    timeout_thread.start()
                    cleanup.append(
                        lambda: cancel.set() and timeout_thread.join()
                    )
                case Limit.Type.MEMORY:
                    previous_soft, _ = resource.getrlimit(resource.RLIMIT_AS)
                    resource.setrlimit(resource.RLIMIT_AS, (limit.value, -1))
                    cleanup.append(
                        lambda: resource.setrlimit(
                            resource.RLIMIT_AS, (previous_soft, -1)
                        )
                    )

        mod_name, func_name = job.func.rsplit(".", 1)
        mod = __import__(mod_name, fromlist=[func_name])
        func = getattr(mod, func_name)

        kwargs = job.kwargs or {}

        try:
            func(**kwargs)
        finally:
            for clean in cleanup:
                clean()

    @staticmethod
    def job_signal_handler(signum: int, frame):
        """
        Handles signals sent to a running job process.

        Subclasses can override this method to provide additional functionality
        or to change the way that signals are handled.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """
        # If we receive a SIGALRM signal, we should raise a TimeoutError, as
        # this is our _TimeoutThread telling us that the job has taken too long
        # to complete.
        if signum == signal.SIGALRM:
            raise TimeoutError("Job timed out.")

    def job_completed(self, future: Future, loop: asyncio.AbstractEventLoop):
        """
        Handles the completion of a job future.

        This method is called when a job future completes, and is responsible
        for cleaning up the job and notifying the queue that the job has
        completed.
        """
        job = self.processes.pop(future)
        exc = future.exception()

        if exc is not None:
            traceback.print_exception(type(exc), exc, exc.__traceback__)

            new_state = dataclasses.replace(
                job,
                state=(
                    JobInstance.State.FAILED
                    if job.attempts + 1 >= job.max_attempts
                    else JobInstance.State.RETRYING
                ),
                attempts=job.attempts + 1,
            )
        else:
            new_state = dataclasses.replace(
                job,
                state=JobInstance.State.SUCCEEDED,
                completed_at=job.started_at,
            )

        f = asyncio.run_coroutine_threadsafe(
            self.queue.push_job_update(new_state),
            loop,
        )
        f.result()



from typing import Callable, Awaitable


class Hub:
    """
    An event hub for registering and emitting events.
    """

    def __init__(self):
        self._handlers: dict[str, list[Callable[..., Awaitable[None]]]] = {}

    def on(self, event: str, f: Callable[..., Awaitable[None]]):
        """
        Register a handler for an event.
        """
        self._handlers.setdefault(event, []).append(f)

    def remove(self, event: str, handler: Callable[..., Awaitable[None]]):
        """
        Remove a handler from an event.
        """
        try:
            self._handlers.get(event, []).remove(handler)
        except ValueError:
            pass

    async def emit(self, event: str, *args, **kwargs):
        """
        Emit an event with the given arguments to all registered handlers.
        """
        for handler in self._handlers.get(event, []):
            await handler(*args, **kwargs)



__all__ = ("Chancy", "Job", "Queue", "Worker", "Limit")

from chancy.app import Chancy
from chancy.executor import Job
from chancy.queue import Queue
from chancy.worker import Worker
from chancy.executor import Limit



import abc
import enum
import asyncio
import typing
import logging
from functools import cached_property

from chancy.logger import logger, PrefixAdapter

if typing.TYPE_CHECKING:
    from chancy.app import Chancy
    from chancy.worker import Worker


class PluginScope(enum.Enum):
    #: The plugin is a general worker plugin.
    WORKER = "worker"
    #: The plugin implements a queue provider.
    QUEUE = "queue"


class Plugin(abc.ABC):
    """
    Base class for all plugins.

    Plugins are used to extend the functionality of the worker. When a worker
    starts, it will call :meth:`run` on all plugins that have a scope that
    matches the worker's scope.

    """

    def __init__(self):
        #: An asyncio.Event that can be used to cancel the plugin.
        self.cancel_signal = asyncio.Event()
        #: An asyncio.Event that can be used to wake up the plugin if it's
        #: sleeping.
        self.wakeup_signal = asyncio.Event()

    async def cancel(self) -> None:
        """
        Cancel the plugin.

        Wakes up the plugin if it's sleeping and attempts to have it exit
        gracefully.
        """
        self.cancel_signal.set()

    @cached_property
    def log(self) -> logging.LoggerAdapter:
        return PrefixAdapter(logger, {"prefix": self.__class__.__name__})

    @classmethod
    @abc.abstractmethod
    def get_scope(cls) -> PluginScope:
        """
        Get the scope of this plugin.
        """

    @abc.abstractmethod
    async def run(self, worker: "Worker", chancy: "Chancy"):
        """
        Runs the plugin.

        This function can and should run indefinitely, as it will be cancelled
        when the worker is stopped.
        """

    async def sleep(self, seconds: int) -> bool:
        """
        Sleep for a specified number of seconds, but allow the plugin to be
        cancelled during the sleep by calling :meth:`cancel`.
        """
        sleep = asyncio.create_task(asyncio.sleep(seconds))
        cancel = asyncio.create_task(self.cancel_signal.wait())
        wakeup = asyncio.create_task(self.wakeup_signal.wait())

        done, pending = await asyncio.wait(
            # As of 3.11, asyncio.wait() no longer accepts coroutines directly,
            # so we must wrap them in asyncio.create_task().
            [sleep, cancel, wakeup],
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()

        if cancel in done:
            raise asyncio.CancelledError()

        self.wakeup_signal.clear()
        return True

    async def wait_for_leader(self, worker: "Worker") -> None:
        """
        Wait until the worker running this plugin is the leader.

        This function will return immediately if the worker is already the
        leader. If the plugin's :meth:`cancel` method is called, this function
        will raise an asyncio.CancelledError.
        """
        cancel = asyncio.create_task(self.cancel_signal.wait())
        leader = asyncio.create_task(worker.is_leader.wait())

        done, pending = await asyncio.wait(
            [cancel, leader],
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()

        if cancel in done:
            raise asyncio.CancelledError()

    def wake_up(self):
        self.wakeup_signal.set()

    def __repr__(self):
        return f"<{self.__class__.__name__}()>"



import asyncio
import json
import uuid
from typing import Any
from asyncio import TaskGroup

from psycopg import sql
from psycopg import AsyncConnection, AsyncCursor

from chancy.app import Chancy
from chancy.hub import Hub
from chancy.plugin import PluginScope
from chancy.logger import PrefixAdapter, logger


class Worker:
    """
    The Worker is responsible for polling queues for new jobs, running any
    configured plugins, and internal management such as heartbeats.

    As an example, lets create a Worker that will poll the "default" queue
    with a concurrency of 10, and run a single worker:

    .. code-block:: python

        async with Chancy(
            dsn="postgresql://localhost/postgres",
            queues=[
                Queue(name="default", concurrency=10),
            ],
        ) as chancy:
            await chancy.migrate()
            await Worker(chancy).start()

    .. note::

        Don't share plugin instances between workers. Each worker should have
        its own instance of each plugin.

    :param chancy: The Chancy application that the worker is associated with.
    :param worker_id: The ID of the worker, which must be globally unique. If
                      not provided, a random UUID will be generated.
    :param heartbeat_poll_interval: The number of seconds between heartbeat
                                    poll intervals.
    :param heartbeat_timeout: The number of seconds before a worker is
                              considered to have lost connection to the
                              cluster.
    """

    def __init__(
        self,
        chancy: Chancy,
        *,
        worker_id: str | None = None,
        heartbeat_poll_interval: int = 30,
        heartbeat_timeout: int = 90,
    ):
        #: The Chancy application that the worker is associated with.
        self.chancy = chancy
        #: The ID of the worker, which must be globally unique.
        self.worker_id = worker_id or str(uuid.uuid4())
        #: The logger used by an active worker.
        self.logger = PrefixAdapter(logger, {"prefix": "Worker"})
        #: The number of seconds between heartbeat poll intervals.
        self.heartbeat_poll_interval = heartbeat_poll_interval
        #: The number of seconds before a worker is considered to have lost
        #: connection to the cluster.
        self.heartbeat_timeout = heartbeat_timeout
        #: An event hub for tracing and debugging.
        self.hub = Hub()
        #: An event that is set when the worker is the leader.
        #: This functionality is not enabled by default - a leadership plugin
        #: must be used to enable this event.
        self.is_leader = asyncio.Event()

    async def start(self):
        """
        Start the worker.

        Will run indefinitely, polling the queues for new jobs and running any
        configured plugins.
        """
        async with self.chancy.pool.connection() as conn:
            self.logger.debug(
                f"Performing initial worker announcement using worker ID"
                f" {self.worker_id!r}."
            )
            await self.announce_worker(conn)

        async with TaskGroup() as group:
            group.create_task(self.maintain_heartbeat())
            group.create_task(self.maintain_notifications())

            for plugin in self.chancy.plugins:
                match plugin.get_scope():
                    case PluginScope.WORKER:
                        group.create_task(plugin.run(self, self.chancy))
                    case PluginScope.QUEUE:
                        group.create_task(plugin.run(self, self.chancy))

    async def maintain_heartbeat(self):
        """
        Announces the worker to the cluster, and maintains a periodic heartbeat
        to ensure that the worker is still alive.

        .. note::

            This method should not be called directly. It is automatically
            run when the worker is started.
        """
        while True:
            async with self.chancy.pool.connection() as conn:
                async with conn.transaction():
                    self.logger.debug("Announcing worker to the cluster.")
                    await self.announce_worker(conn)
            await asyncio.sleep(self.heartbeat_poll_interval)

    async def maintain_notifications(self):
        """
        Listen for notifications from the database.

        Improves the reactivity of a worker by allowing it to almost immediately
        react to database events using Postgres's LISTEN/NOTIFY feature.

        .. note::

            This feature utilizes a permanent connection to the database
            separate from the shared connection pool.

        .. note::

            This method should not be called directly. It is automatically
            run when the worker is started.
        """
        connection = await AsyncConnection.connect(
            self.chancy.dsn, autocommit=True
        )
        await connection.execute(
            sql.SQL("LISTEN {channel};").format(
                channel=sql.Identifier(f"{self.chancy.prefix}events")
            )
        )
        self.logger.info("Started listening for realtime notifications.")
        async for notification in connection.notifies():
            j = json.loads(notification.payload)
            match j["t"]:
                case "pushed":
                    try:
                        await self[j["q"]].wake_up()
                    except KeyError:
                        # This worker isn't setup to process this queue, so
                        # we'll just ignore it.
                        continue
                case _:
                    await self.hub.emit(j["t"], j)

    async def announce_worker(self, conn: AsyncConnection):
        """
        Announce the worker to the cluster.

        This will insert the worker into the workers table, or update the
        `last_seen` timestamp if the worker is already present.

        :param conn: The connection to use for the announcement.
        """
        async with conn.cursor() as cur:
            await cur.execute(
                sql.SQL(
                    """
                    INSERT INTO {workers}
                        (worker_id, last_seen)
                    VALUES (%s, NOW())
                    ON CONFLICT (worker_id) DO UPDATE
                        SET last_seen = NOW()
                    """
                ).format(
                    workers=sql.Identifier(f"{self.chancy.prefix}workers")
                ),
                [self.worker_id],
            )
            await self.notify(
                cur, "worker.started", {"worker_id": self.worker_id}
            )

        await self.hub.emit("worker.started")

    async def revoke_worker(self, conn: AsyncConnection):
        """
        Revoke the worker from the cluster.

        This will remove the worker from the workers table, making it appear
        as though the worker has gone offline.

        :param conn: The connection to use for the revocation.
        """
        async with conn.cursor() as cur:
            await cur.execute(
                sql.SQL(
                    """
                    DELETE FROM {workers}
                    WHERE worker_id = %s
                    """
                ).format(
                    workers=sql.Identifier(f"{self.chancy.prefix}workers")
                ),
                [self.worker_id],
            )
            await self.notify(
                cur, "worker.stopped", {"worker_id": self.worker_id}
            )

        await self.hub.emit("worker.stopped")

    async def notify(
        self, cursor: AsyncCursor, event: str, payload: dict[str, Any]
    ):
        """
        Notify the cluster of an event.

        .. note::

            This method does not start or end a transaction. It is up to the
            caller to manage the transaction.

        :param cursor: The cursor to use for the notification.
        :param event: The event to notify the cluster of.
        :param payload: The payload to send with the notification.
        """
        await cursor.execute(
            sql.SQL(
                """
                SELECT pg_notify(%s, %s)
                """
            ),
            [
                f"{self.chancy.prefix}events",
                json.dumps({"t": event, **payload}),
            ],
        )

    def __repr__(self):
        return f"<Worker({self.worker_id!r})>"



import logging


logger = logging.getLogger("chancy")
logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setFormatter(
    logging.Formatter(
        fmt="%(asctime)s • %(levelname)s • %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
)
logger.addHandler(handler)


class PrefixAdapter(logging.LoggerAdapter):
    """
    A logging adapter that prefixes log messages with a custom string.
    """

    def process(self, msg, kwargs):
        return f"{self.extra['prefix']} • {msg}", kwargs



"""
Utilities for creating dynamic rules that can be used when configuring the
conditions of a Plugin.
"""

import abc
from typing import Union

from psycopg import sql

OpT = Union["Conditional", "Operator"]


class SQLAble(abc.ABC):
    @abc.abstractmethod
    def to_sql(self) -> sql.Composed:
        """
        Convert this object to an SQL expression.
        """


class Operator(SQLAble, abc.ABC):
    def __init__(self, left: OpT, right: OpT):
        self.left = left
        self.right = right

    def __or__(self, other: OpT) -> "Or":
        return Or(self, other)

    def __add__(self, other: OpT) -> "And":
        return And(self, other)


class Or(Operator):
    def to_sql(self) -> sql.Composed:
        return sql.SQL("({left}) OR ({right})").format(
            left=self.left.to_sql(), right=self.right.to_sql()
        )


class And(Operator):
    def to_sql(self) -> sql.Composed:
        return sql.SQL("({left}) AND ({right})").format(
            left=self.left.to_sql(), right=self.right.to_sql()
        )


class Conditional(SQLAble, abc.ABC):
    """
    A conditional is a rule that can be applied to a pruner.
    """

    def __or__(self, other: OpT) -> Or:
        return Or(self, other)

    def __add__(self, other: OpT) -> And:
        return And(self, other)


class QueueRule(Conditional):
    """
    A conditional that matches jobs based on their queue.
    """

    def __init__(self, queue: str):
        self.queue = queue

    def to_sql(self) -> sql.Composed:
        return sql.SQL("queue = {queue}").format(queue=sql.Literal(self.queue))


class AgeRule(Conditional):
    """
    A conditional that matches jobs based on their age.
    """

    def __init__(self, age: int):
        self.age = age

    def to_sql(self) -> sql.Composed:
        return sql.SQL(
            "scheduled_at < NOW() - INTERVAL '{age} seconds'"
        ).format(age=sql.Literal(self.age))


class JobRule(Conditional):
    """
    A conditional that matches jobs based on their job name.
    """

    def __init__(self, job: str):
        self.job = job

    def to_sql(self) -> sql.Composed:
        return sql.SQL("payload->>'func' = {job}").format(
            job=sql.Literal(self.job)
        )


RuleT = Union[SQLAble, list[SQLAble]]



import dataclasses
from functools import cached_property

from psycopg_pool import AsyncConnectionPool

from chancy.migrate import Migrator
from chancy.queue import Queue
from chancy.plugin import Plugin


@dataclasses.dataclass(kw_only=True)
class Chancy:
    """
    The main application object for Chancy.

    This object is responsible for managing the connection pool to the database,
    general configuration, and provides some helpers for common tasks like
    pushing jobs.
    """

    #: The DSN to use to connect to the database.
    dsn: str
    #: The list of plugins that are to be used by the Chancy application.
    plugins: list[Plugin] = dataclasses.field(default_factory=list)
    #: The prefix to use for all Chancy database tables.
    prefix: str = "chancy_"
    #: The minimum number of connections to keep in the connection pool.
    min_connection_pool_size: int = 1
    #: The maximum number of connections to keep in the connection pool.
    max_connection_pool_size: int = 10
    #: The number of seconds to wait before attempting to reconnect to the
    #: database after a connection is lost from the pool.
    poll_reconnect_timeout: int = 60 * 5

    def __post_init__(self):
        # Ensure there are no queues with duplicate names
        queue_names = [p.name for p in self.plugins if isinstance(p, Queue)]
        if len(queue_names) != len(set(queue_names)):
            raise ValueError("Duplicate queue names are not allowed.")

    async def migrate(self, *, to_version: int | None = None):
        """
        Migrate the database to the latest schema version.

        If `to_version` is provided, the database will be migrated to that
        specific version, up or down as necessary.
        """
        migrator = Migrator("chancy", "chancy.migrations", prefix=self.prefix)
        async with self.pool.connection() as conn:
            await migrator.migrate(conn, to_version=to_version)

    @cached_property
    def pool(self):
        """
        An async connection pool to the configured database.
        """
        return AsyncConnectionPool(
            self.dsn,
            open=False,
            check=AsyncConnectionPool.check_connection,
            min_size=self.min_connection_pool_size,
            max_size=self.max_connection_pool_size,
            reconnect_timeout=self.poll_reconnect_timeout,
        )

    async def __aenter__(self):
        await self.pool.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.pool.close()
        return False

    async def open(self):
        """
        Open the connection pool.

        Whenever possible, it's preferred to use the async context manager
        instead of this method, as it will ensure that the connection pool is
        properly closed when the block is exited. Ex:

        .. code-block:: python

                async with Chancy(...) as chancy:
                    ...

        """
        await self.pool.open()

    async def close(self):
        """
        Close the connection pool.

        Whenever possible, it's preferred to use the async context manager
        instead of this method, as it will ensure that the connection pool is
        properly closed when the block is exited. Ex:

        .. code-block:: python

                    async with Chancy(...) as chancy:
                        ...
        """
        if not self.pool.closed:
            await self.pool.close()

    async def push(self, queue: str | Queue, *jobs):
        """
        Push one or more jobs onto a queue.

        :param queue: The name or Queue of the queue to push the job onto.
        :param jobs: The jobs to push onto the queue.
        """
        if isinstance(queue, Queue):
            await queue.push(self, list(jobs))
            return

        await self[queue].push(self, list(jobs))

    @cached_property
    def queues(self):
        """
        A list of all queues that are managed by the Chancy application.
        """
        return {p.name: p for p in self.plugins if isinstance(p, Queue)}

    def __getitem__(self, queue_name: str):
        return self.queues[queue_name]
